import sys, ast
import copy
import time
import collections
from da.compiler.dast import *
from da.compiler.ui import daast_from_file
import dagen
import da.compiler.dast

config_params = {}
auxiliary_configs = ['termination_check' , 'source', 'destination', 'main', 'send_transformation', 'send_coordination_bound', 'approx_coordination', 'equality_check']
# string representations
my_send_str = 'my_send'
send_str = 'send'
new_str = 'new'
gateway_str = 'Gateway_'
module_names = set()
# generate ast of hepler methods and statements from helper file
ast_helper = daast_from_file('HelperMethods.py')
imports_added = []
def transformDast(dast):
    # a list to keep track of the nodes processed
    memo = []
    # if process being transformed is the main process
    # if node is of type program, check for entry point (main method) and call transform_node on the entry point
    if isinstance(dast, Program):
        if hasattr(dast, 'nodecls') and hasattr(dast.nodecls, 'entry_point'):
            # print('program has attrib nodecls : ', type(node.nodecls.entry_point))
            transform_main(dast, memo)
        transform_node(dast, memo)


def transform_main(node, memo):
    global imports_added

    # modify 'main' method
    for treenode in ast.walk(node.nodecls):
        if treenode not in memo:
            # replace calls to new( orig_process_name, ...) with new( Gateway_orig_process_name, ...)
            if isinstance(treenode, BuiltinCallExpr) and treenode.subexprs[0] == new_str:
                modify_new(treenode)
            memo.append(treenode)
    # add Gateway_CLASS import statements (we do not modify the original import statements as the main mathod may need APIs from those modules)
    if isinstance(node, Program):
        node_body = []
        for key, val in config_params.items():
            if key not in auxiliary_configs and key+'.da' not in module_names:
                import_stmt = ImportStmt(parent = node)
                import_stmt.items.append(Alias(import_stmt, name=gateway_str +  key))
                node_body.append(import_stmt)
                imports_added.append(gateway_str + '_'+ key)
        node_body.append(node.body)
        node.body = node_body

# recursive method to transform DAST
def transform_node(node, memo):
    global config_params, imports_added, ast_helper
    if node not in memo:
        # replace calls to new( orig_process_name, ...) with new( Gateway_orig_process_name, ...)
        if isinstance(node, BuiltinCallExpr) and node.subexprs[0] == new_str:
            modify_new(node)
        if isinstance(node, ImportStmt):
            # print(" IMPORt stmt ",node.items)
            for i in node.items:
                if isinstance(i.name,NamedVar) and i.name.name in config_params:
                    if gateway_str + i.name.name in imports_added:
                        orig_import = ImportStmt(parent=node)
                        alias = Alias(orig_import)
                        namedvar = NamedVar(name=i.name.name)

                        namedvar.parent = alias
                        alias.name = namedvar
                        orig_import.items.append(alias)

                        node.body = [orig_import] + node.body
                elif isinstance(i.name, str) and i.name  in config_params:
                    # print(' FOUND name to replac in import stsmti.name : ',i.name)
                    if gateway_str + i.name in imports_added:
                        orig_import = ImportStmt(parent=node)
                        alias = Alias(orig_import, name = i.name)
                        orig_import.items.append(alias)
                        node.body = [orig_import] + node.body
        # if node is an instance of Process, insert 'gateway' arg in setup method
        # and add subtrees of definitions of 'my_send' and 'replace' methods to the DAST.
        if isinstance(node, Process):
            if node.name not in config_params:
                memo.append(node)
                return
            if hasattr(node,'setup'):
                add_gateway_arg(node)
            if hasattr(node, 'events'):
                for evt in node.events:
                    transform_node(evt.handlers, memo)
            add_helper_methods(node)

        # replace all instances of  'send' to 'my_send' method
        if isinstance(node, SimpleStmt) and hasattr(node,'expr') and isinstance(node.expr, BuiltinCallExpr) and node.expr.subexprs[0] == send_str:
            node.expr.subexprs[0] = my_send_str

        # check if node has body attribute, if yes, check if it is a collection or single object
        if hasattr(node, 'body'):
            if isinstance(node.body, collections.Iterable):
                for child in node.body:
                    transform_node(child, memo)
            else:
                transform_node(node.body, memo)

        if hasattr(node, 'value'):
            # print('\nvalue : ', node.value)
            transform_node(node.value, memo)

        if hasattr(node, 'targets'):
            # print('\n node.targets : ', node.targets)
            transform_node(node.targets, memo)

        # if the node represents a function (CallExpr), transform the function and then its arguments
        if hasattr(node, 'func'):
            transform_node(node.func, memo )
            if hasattr(node, 'args'):
                transform_node(node.args, memo)

        # some node attributes are lists which contain references to the child nodes
        # traverse the list for inserting transformation
        if isinstance(node, list):
            for n in node:
                transform_node(n, memo)

        if isinstance(node, IfStmt):
            if node.elsebody:
                if isinstance(node.elsebody, collections.Iterable):
                    for child in node.elsebody:
                        transform_node(child, memo)
                else:
                    transform_node(node.elsebody, memo)

        if isinstance(node, Statement) and node.label and not isinstance(node, AwaitStmt):
            yield_without_await = get_yield_without_await_ast(node)
            node.parent.replace_child(node,yield_without_await)
            #  node.expr contains the next statement after the label
            # this statement needs to be added as the next child to parent of node after the yield_without_await
            parent_body = []
            # print("Parent body before : ", node.parent.body)
            next_expr = None
            next_expr = get_next_expr(node)
            for item in node.parent.body:
                parent_body.append(item)
                if item == yield_without_await and next_expr:
                    parent_body.append(next_expr)
            if parent_body:
                node.parent.body = parent_body
            # print("Parent body after : ", node.parent.body)
        # replace simple 'await(<condition>)' statement with 'while not(condition): my_yield(True,None)'
        if isinstance(node, AwaitStmt):
            if (len(node.branches) == 1 or not node.branches[0].body):
                replace_await = get_replace_simple_await(node)
                # replace the awaitstmt node with whilestmt node
                node.parent.replace_child(node, replace_await)
            # await with branches but no timeout
            elif node.timeout == None and len(node.branches) > 1:
                replace_await = get_replace_await_multiple_branches(node)
                # replace the awaitstmt node with whilestmt node
                node.parent.replace_child(node, replace_await)
                # generate the if-else for branches
                if_else_for_branches = get_if_else_for_branches(node, replace_await)
                transform_node(if_else_for_branches, memo)
                parent_body_list = []
                for item in node.parent.body:
                    parent_body_list.append(item)
                    if item == replace_await:
                        parent_body_list.append(if_else_for_branches)
                if parent_body_list:
                    node.parent.body = parent_body_list

            # await statements with timeout
            elif node.timeout and len(node.branches) > 1:
                replace_await = get_replace_await_multiple_branches_timeout(node)
                # replace the awaitstmt node with whilestmt node
                node.parent.replace_child(node, replace_await)
                # generate the if-else for branches with timeout
                if_else_for_branches = get_if_else_for_branches_timeout(node, replace_await)
                transform_node(if_else_for_branches, memo)
                parent_body_list = []
                for item in node.parent.body:
                    if item == replace_await:
                        # add statement : start_time = time.time() from ast helper
                        parent_body_list.append(ast_helper.body[8])
                        parent_body_list.append(replace_await)
                        parent_body_list.append(if_else_for_branches)
                    else:
                        parent_body_list.append(item)
                if parent_body_list:
                    node.parent.body = parent_body_list
                    print("timeout : ", parent_body_list)
        memo.append(node)

def get_yield_without_await_ast(node):
    global ast_helper
    yield_without_await = SimpleStmt(parent=node.parent)
    # create a list for expression as label is replaced with a CallExpr which is bound to the next statement
    # the next statement is in the expr field of this node with label
    yield_without_await.expr = []
    for k, v in ast_helper.body[5].__dict__.items():
        if k == 'expr':
            yield_without_await.expr += [v]
            continue
        if not k == '_parent':
            setattr(yield_without_await, k, v)
    return yield_without_await

def get_next_expr(node):
    new_expr = SimpleStmt(parent=node.parent)
    if hasattr(node, 'expr') and node.expr:
        type_expr = type(node.expr)
        expr = type_expr(parent=node.parent)
        for k, v in node.expr.__dict__.items():
            if not k == '_parent':
                setattr(expr, k, v)
    else:
        type_expr = type(node)
        new_expr = SimpleStmt(parent=node.parent)
        expr = type_expr(parent=new_expr)
        for k, v in node.__dict__.items():
            if not k == '_parent':
                setattr(expr, k, v)
    setattr(expr, 'label', None)
    setattr(new_expr, 'expr', expr)
    setattr(new_expr, 'label', None)
    return new_expr


def get_if_else_for_branches(node, new_node):
    first = True
    first_if_stmt = IfStmt(parent=new_node.parent)
    for branch in node.branches:
        if branch is not None:
            if first:
                first_if_stmt.condition = branch.condition
                first_if_stmt.body = []
                for item in  branch.body:
                    t = type(item)
                    new_item = t(parent=first_if_stmt)
                    for k,v in item.__dict__.items():
                        if not k == '_parent':
                            setattr(new_item,k,v)
                    first_if_stmt.body.append(new_item)
                if_stmt = first_if_stmt
                first = False
            else:
                temp_if = IfStmt(parent=if_stmt)
                temp_if.condition = branch.condition
                temp_if.body = []
                for item in branch.body:
                    t = type(item)
                    new_item = t(parent=if_stmt)
                    for k, v in item.__dict__.items():
                        if not k == '_parent':
                            setattr(new_item, k, v)
                    temp_if.body.append(new_item)
                if_stmt.elsebody.append(temp_if)
                if_stmt = temp_if
    # add else: assert(false)
    final_else = AssertStmt(parent=if_stmt)
    final_else.expr = FalseExpr(parent = final_else)
    if_stmt.elsebody.append(final_else)
    return first_if_stmt

def get_if_else_for_branches_timeout(node, new_node):
    global ast_helper

    # # add import stmt for 'time' module if missing
    # add_time_import(node)

    first = True
    first_if_stmt = IfStmt(parent=new_node.parent)
    for branch in node.branches:
        if branch.condition is not None:
            if first:
                first_if_stmt.condition = branch.condition
                first_if_stmt.body = []
                for item in branch.body:
                    t = type(item)
                    new_item = t(parent= first_if_stmt)
                    for k,v in item.__dict__.items():
                        if not k == '_parent':
                            setattr(new_item,k,v)
                    first_if_stmt.body.append(new_item)
                if_stmt = first_if_stmt
                first = False
            else:
                temp_if = IfStmt(parent = if_stmt)
                temp_if.condition = branch.condition
                temp_if.body = []
                for item in branch.body:
                    t = type(item)
                    new_item = t(parent= if_stmt)
                    for k,v in item.__dict__.items():
                        if not k == '_parent':
                            setattr(new_item,k,v)
                    temp_if.body.append(new_item)
                if_stmt.elsebody.append(temp_if)
                if_stmt = temp_if
        else:
            if_stmt.elsebody = []
            for item in branch.body:
                t = type(item)
                new_item = t(parent=if_stmt)
                for k, v in item.__dict__.items():
                    if not k == '_parent':
                        setattr(new_item, k, v)
                if_stmt.elsebody.append(new_item)
    return first_if_stmt

def get_replace_await_multiple_branches(node):
    global ast_helper
    replace_await_stmt = WhileStmt(parent=node.parent)
    for k, v in ast_helper.body[7].__dict__.items():
        if k == 'condition':
            logical_expr = LogicalExpr(parent=replace_await_stmt, op=NotOp)
            inner_logical_expr = LogicalExpr(parent=logical_expr, op=OrOp)
            logical_expr.subexprs = [inner_logical_expr]
            inner_logical_expr.subexprs = []
            for branch in node.branches:
               if branch.condition is not None:
                    inner_logical_expr.subexprs.append(branch.condition)
            setattr(replace_await_stmt, k, logical_expr)
            continue
        if not k == '_parent':
            setattr(replace_await_stmt, k, v)
    return replace_await_stmt


def get_replace_await_multiple_branches_timeout(node):
    global ast_helper
    replace_await_stmt = WhileStmt(parent=node.parent)
    for k, v in ast_helper.body[9].__dict__.items():
        if k == 'condition':
            logical_expr = LogicalExpr(parent=replace_await_stmt, op=NotOp)
            inner_logical_expr = LogicalExpr(parent=logical_expr, op=OrOp)
            logical_expr.subexprs = [inner_logical_expr]
            inner_logical_expr.subexprs = []
            for branch in node.branches:
                if branch.condition is not None:
                    inner_logical_expr.subexprs.append(branch.condition)
            setattr(replace_await_stmt, k, logical_expr)
            continue
        if not k == '_parent':
            setattr(replace_await_stmt, k, v)
    # replace value of timeout into 't' in the Helper ast for await with timeout
    setattr(ast_helper.body[8].body[1].value, 'left', node.timeout)
    return replace_await_stmt

def get_replace_simple_await(node):
    global ast_helper
    replace_await_stmt = WhileStmt(parent = node.parent)
    for k, v in ast_helper.body[6].__dict__.items():
        if k == 'condition':
            logical_expr = LogicalExpr(parent = replace_await_stmt, op=NotOp, subexprs= [node.branches[0].condition])
            setattr(replace_await_stmt,k,logical_expr)
            continue
        if not k == '_parent':
            setattr(replace_await_stmt, k, v)
    return replace_await_stmt


def add_gateway_arg(node):
    print(' adding gateway arg: ', node.name)
    gateway = [NamedVar(parent=copy.copy(node.setup.parent), name='gateway')]
    node.args.args = gateway + node.args.args


# modify occurrences of new
def modify_new(node):
    # process arg to new is name of the process
    # if it is an AttributeExpr, replace name and attribute with Gateway_<NAME> if they are names of modules in config file
    if(isinstance(node.subexprs[1][0], AttributeExpr) and hasattr(node.subexprs[1][0].value.subexprs[0],'name')):
        orig_name = node.subexprs[1][0].value.subexprs[0].name
        if orig_name in config_params:
            node.subexprs[1][0].value.subexprs[0].name = gateway_str + node.subexprs[1][0].value.subexprs[0].name
        if node.subexprs[1][0].attr in config_params:
            node.subexprs[1][0].attr =  gateway_str + node.subexprs[1][0].attr

    # if it is a NamedVar, replace its name with Gateway_<NAME> if it occurs in config file
    elif isinstance(node.subexprs[1][0], SimpleExpr) and isinstance(node.subexprs[1][0].value, NamedVar):
        if node.subexprs[1][0].value.name in config_params:
            node.subexprs[1][0].value.name = gateway_str + node.subexprs[1][0].value.name + "." + gateway_str + node.subexprs[1][0].value.name

def add_helper_methods(process):
    global ast_helper
    add_copy_import(process)
    add_time_import(process)
    my_send = Function(parent = process)
    if 'send_transformation' in config_params and config_params['send_transformation'] == 'True':
        for k,v in ast_helper.body[0].__dict__.items():
            if not isinstance(v, Program):
                setattr(my_send,k,v)
    else:
        for k,v in ast_helper.body[1].__dict__.items():
            if not isinstance(v, Program):
                setattr(my_send,k,v)
    process.methods.append(my_send)

    replace_func = Function(parent=process)
    for k, v in ast_helper.body[2].__dict__.items():
        if not isinstance(v, Program):
            setattr(replace_func, k, v)
    process.methods.append(replace_func)

    add_yield_num_field(process, ast_helper.body[3])

    my_yield_func = Function(parent=process)
    for k, v in ast_helper.body[4].__dict__.items():
        if not isinstance(v, Program):
            setattr(my_yield_func, k, v)
    for func_node in ast.walk(my_yield_func):
        # The parser adds a label on the await statement, we remove it
        if isinstance(func_node, AwaitStmt):
            setattr(func_node, 'label', "")
    process.methods.append(my_yield_func)


def  add_copy_import(process):
    for item in process.parent.body:
        if isinstance(item, ImportStmt):
            for i in item.items:
                if i.name == 'copy':
                    return
        elif isinstance(item, ImportFromStmt) and item.module == 'copy':
            for i in item.items:
                if i.name == 'deepcopy':
                    return
    importstmt = ImportStmt(parent = process.parent)
    importstmt.items.append(NamedVar(name = 'copy'))
    process.parent.body = [importstmt] + process.parent.body

def add_time_import(process):
    for item in process.parent.body:
        if isinstance(item, ImportStmt):
            for i in item.items:
                if i.name == 'time':
                    return
        elif isinstance(item, ImportFromStmt) and item.module == 'time':
            for i in item.items:
                if i.name == 'time':
                    return
    importstmt = ImportStmt(parent = process.parent)
    importstmt.items.append(NamedVar(name = 'time'))
    process.parent.body = [importstmt] + process.parent.body

def add_yield_num_field(node, assignment_stmt):
    assignment_yeild_num = AssignmentStmt(parent = node.setup.body)
    for k, v in assignment_stmt.__dict__.items():
        setattr(assignment_yeild_num, k, v)
    setup_body = [assignment_yeild_num] + node.setup.body
    node.setup.body = setup_body


def replace_yield_without_await(node):
    global ast_helper
    yield_without_await = Statement(parent = node.parent)
    for k, v in ast_helper.body[5].__dict__.items():
        if not k == '_parent':
            setattr(yield_without_await, k, v)
    node = ast.copy_location(yield_without_await, node)
    return node


def transform_gateway(f):
    global config_params
    with open(f, 'r') as myfile:
        gw_code = myfile.read()
    diversified_class = []
    for key, val in config_params.items():
        if key  not in auxiliary_configs and key+'.da' not in module_names:
            diversified_class.append(key)

    # print('diversified_class list : ', diversified_class)
    for cls in diversified_class:
        create_gateway(gw_code, cls)


def create_gateway(gw_code, mod):
    global config_params
    module_gw = gw_code.replace('CLASS', mod)
    dest_location = config_params['destination']
    with open(dest_location + "\\" + "Gateway_" + mod + ".da", "w+") as f:
        f.write(module_gw)

# read config file and get dictionary of config parameters
def read_config(config_file):
    global module_names
    config_params = {}
    with open(config_file, 'r') as f:
        for line in f:
            if line[0] != '#':
                (key, sep, val) = line.partition('=')
                # if the line does not contain '=', it is invalid and hence
                # ignored
                if len(sep) != 0:
                    if key.strip() == 'source':
                        val = val.strip()
                        config_params['source'] =  val
                    elif key.strip() == 'destination':
                        val = val.strip()
                        config_params['destination'] =  val
                    elif key.strip() == 'termination_check':
                        val = val.strip()
                        config_params['termination_check'] = int(val) if str.isdecimal(val) else val
                    elif key.strip() == 'main':
                        val = val.strip()
                        config_params['main'] =  val
                    else:
                        value = val.split(':')
                        # value contains class : module1, module2, ...
                        if len(value) > 1:
                            modules = value[1].split(',')
                            config_params[value[0].strip()] = []
                            for m in modules:
                                config_params[value[0].strip()].append(m.strip())
                                module_names.add(m.strip() + '.da')
                        else:
                            val = val.strip()
                            config_params[key.strip()] = int(val) if str.isdecimal(val) else val
    module_names.add('Gateway.da')
    module_names.add(config_params['main']+'.da')

    print('\n\nconfig_params : ',config_params,'\n\n')
    return config_params

def main():
    global config_params, module_names
    config_file = sys.argv[1]
    config_params = read_config(config_file)
    src_location = config_params['source']
    dest_location = config_params['destination']
    print('module_names : ', module_names )
    for f in module_names:
        try:
            # transform Gateway code : only replace occurrences of  'CLASS with '
            if "Gateway.da" in f:
                print("\nTRansforming gateway")
                transform_gateway(f)
            else:
                try:
                    print("\n Transforming ", f)
                    astree = daast_from_file(src_location+ "\\" + f)
                    transformDast(astree)
                    # print('************************************************************************************')
                    #print(dagen.to_pseudo(astree))
                except Exception as e:
                    print('Something went wrong ', e)
                with open(dest_location + "\\" + f, "w+") as f:
                    print('writing file to ', dest_location)
                    f.write(dagen.to_pseudo(astree))
                    # print(dagen.to_pseudo(astree))
        except Exception as e:
            print(e)

if __name__ == '__main__':
    main()